# Licensed under a 3-clause BSD style license - see LICENSE.rst

# The astropy affiliated package template uses this directory for any python
# packages that are bundled with the affiliated package but are external to the
# affiliated package, and hence are developed in a separate source tree. Note
# that this package is distinct from the /cextern directory of the source code
# distribution, as that directory only contains C extension code.
