# Autogenerated by Astropy-affiliated package speclite's setup.py on 2021-09-09 18:39:12 UTC
import datetime

version = "0.14"
githash = "7c4e59fd0e217fcb6ad8ba1146cb0ab14fa3e875"


major = 0
minor = 14
bugfix = 0

version_info = (major, minor, bugfix)

release = True
timestamp = datetime.datetime(2021, 9, 9, 18, 39, 12)
debug = False

astropy_helpers_version = "4.0.1"
